import pandas as pd
import numpy as np
import random
import json
from celibration import (
    CalibrationModel,
    CalibrationModelFactory,
    CalibrationDifferenceFunction,
    CalibrationModelReport,
)
from typing import Any

from celibration.plugins.calibrationmodels.genetic_algorithm.problem import (
    GAProblem,
)
from celibration.plugins.calibrationmodels.genetic_algorithm.callback_convergence import (
    CallBack
)

from stylized_supply_chain_model.RunnerObject import  (
    SupplyChainSimModel,
)

from ema_workbench.em_framework.optimization import NSGAII, EpsNSGAII
from platypus import nondominated, CompoundOperator, SBX, HUX, PM, BitFlip

class PluginFactory(CalibrationModelFactory):
    def create_calibration_model(self, info=dict) -> CalibrationModel:
        return GeneticAlgorithmModel(info=info)


class GeneticAlgorithmModelReport(CalibrationModelReport):
    # self._model = model is set in the default constructor
    def get_string(self) -> str:
        missing_values = self._model.pluginscope["df_in"].isnull().sum().sum()
        result = (
            f"Report:\033[1m GeneticAlgorithmModelReport \033[0m:\n"
            # f"\tParameters:\n"
            # f"\t\t{self._kwargs}\n"
            f"\tModel:\n"
            f"\tDf_in missing: {missing_values} \n"
            f"\t\tResults: {self._model.get_summary()[1]}\n"
        ).expandtabs(2)
        return result

    def render(self, type: str) -> Any:
        return NotImplementedError

    def export_to_file(self, type: str, filename: str) -> Any:
        if type == "pdf":
            self.export_to_pdf(filename=filename)
        if type == "json":
            self.export_to_json(filename=filename)
        return NotImplementedError

    def export_to_json(self, filename: str) -> Any:
        cal_model_json = {
            "Model_info": self._model.get_info(),
            "Score": self._model.pluginscope["score"],
            "Results": self._model.pluginscope["solutions"],
            "Convergence": {"nfe": self._model.pluginscope["nfe_list"],
                            "eprogress": self._model.pluginscope["eprogress"]}
        }
        with open(filename, "w", encoding="utf-8") as f:
            f.write(json.dumps(cal_model_json))


class GeneticAlgorithmModel(CalibrationModel):
    def fit(
        self,
        df_in: pd.DataFrame,
        diff_func_object: CalibrationDifferenceFunction,
        diff_func_parameters: dict,
        debug: bool,
        **kwargs,
    ):
        """Performs the calibration via a Genetic Algorithm (GA).
        First, the problem is initialized. It can be multiobjective or not. Default is mono-objective.
        Next, the optimization is runned by an algorithm. Default algorithm used by GA is NSGAII.
        Other alternatives are epsNSGAII or BORG. It returns one optimal solution (in the case of mono-objective),
        or a Pareto front (in the case of multi-objective).

        Args:
            df_in (Dataframe): dataframe to fit on
            diff_func_object (object): difference function to use as likelihood
            kwargs: arguments in key-value format

        Returns:
            None
        """
        if "seed" in kwargs:
            random.seed(kwargs["seed"])
            np.random.seed(kwargs["seed"])

        if "multi_objective" in kwargs:
            multi_objective = kwargs["multi_objective"]
        else:
            multi_objective = False

        self.pluginscope["decision_variables"] = kwargs["decision_variables"]

        problem = self.initialize_problem(
            decision_variables=kwargs["decision_variables"],
            ranges_variables=kwargs["ranges_variables"],
            obj_names=kwargs["objectives"],
            simulation_model=SupplyChainSimModel,
            distance_metrics=diff_func_object,
            df_ground_truth=df_in,
            multi_objective=multi_objective,
        )

        variator = CompoundOperator(SBX(), HUX(), PM()) #BitFlip() does not work without integer
        algorithm = NSGAII(
            problem,
            population_size=kwargs["population_size"],
            variator=variator,
        )

        if "algorithm" in kwargs:
            if kwargs["algorithm"] == "NSGAII":
                pass
            if kwargs["algorithm"].upper() == "EPSNSGAII":
                algorithm = EpsNSGAII(
                    problem,
                    epsilons=kwargs["epsilons"],
                    population_size=kwargs["population_size"],
                    variator=variator,
                )
        # nfe = population_size * generation
        callback = CallBack()
        algorithm.run(kwargs["nfe"], callback)

        solutions = self.get_solutions(algorithm)
        self.pluginscope["solutions"] = solutions
        # use pluginscope to use 'self' variables in other functions
        self.pluginscope["score"] = self.calculate_score(
            solutions, kwargs["objectives"], kwargs["ranges_variables"]
        )

        self.pluginscope["df_in"] = df_in

        self.pluginscope["nfe_list"] = callback.nfe
        self.pluginscope["eprogress"] = callback.eprogress

        return

    def initialize_problem(
        self,
        decision_variables: list,
        ranges_variables: list,
        obj_names: list,
        simulation_model: object,
        distance_metrics: CalibrationDifferenceFunction,
        df_ground_truth: pd.DataFrame,
        multi_objective: bool,
    ):
        """Initializes the problem.

        Args:
            decision_variables (list): list of decision variables
            ranges_variables (list): list of min and max value that a decision variable can take
            obj_names (list): list of the names of the objective columns
            simulation_model (object): Simulation Model Object (DSOL)
            distance_metrics (object): difference function
            df_ground_truth (Dataframe): df_in
            multi_objective (bool): True or False

        Returns:
            problem (object)
        """
        problem = GAProblem(
            decision_variables,
            ranges_variables,
            obj_names,
            simulation_model,
            distance_metrics,
            df_ground_truth,
            multi_objective,
        )
        return problem

    def get_solutions(self, algorithm):
        """Transforms the solutions from the solution of GA to a readible values. It represents
        the decision variables of the solution and the objective of the nondominated solutions.

        Args:
            algorithm (object): algorithm of GA including results

        Returns:
            dict_all_solutions (dict)
        """
        dict_all_solutions = {}

        nondominated_solutions = nondominated(algorithm.result)

        for i in range(len(nondominated_solutions)):
            solution = nondominated_solutions[i]
            dict_all_solutions[i] = {}
            for var in range(algorithm.problem.nvars):
                decision_variable_name = algorithm.problem.decision_variables[var]
                var_value = algorithm.problem.types[var].decode(solution.variables[var])
                dict_all_solutions[i][decision_variable_name] = var_value

            for obj in range(algorithm.problem.nobjs):
                objective_name = algorithm.problem.obj_names[obj]
                obj_value = solution.objectives[obj]
                dict_all_solutions[i][objective_name] = obj_value

        return dict_all_solutions

    def calculate_score(self, solutions: dict, objectives: list, ranges: list):
        """Calculates performance of the model and difference function.
        [extend this]

        Args:
            solutions (dict): dictionairy of all solutions
            objectives (list): names of the objectives

        Returns:
            a single score
        """
        all_objectives = [
            value
            for i in range(len(solutions))
            for key, value in solutions[i].items()
            if key in objectives
        ]

        df_solutions = pd.DataFrame.from_dict(solutions, orient="index")

        if all(x == all_objectives[0] for x in all_objectives):
            return self.normalize_solution_score(df_solutions, ranges)
            # return all_objectives[0]
        else:
            all_scores_normalized = []
            for i in range(len(df_solutions)):
                score = self.normalize_solution_score(df_solutions.iloc[i], ranges)
                all_scores_normalized.append(score)
            return sum(all_scores_normalized) / len(all_scores_normalized)
            # return sum(all_objectives) / len(all_objectives)

    def normalize_solution_score(self, solution: pd.DataFrame, ranges: list):
        # TODO generiek maken, ground truth parameters & aantal parameters
        """How close to the ground truth parameters"""
        min_par1 = ranges[0][0]
        max_par1 = ranges[0][1]
        # min_par2 = ranges[1][0]
        # max_par2 = ranges[1][1]

        gt_par1 = self.normalize(2.5, min_par1, max_par1)
        # gt_par2 = self.normalize(1, min_par2, max_par2)

        # Normalize the difference
        try:
            var_values = [
                solution[var][0] for var in self.pluginscope["decision_variables"]
            ]
        except IndexError:
            var_values = [
                solution[var] for var in self.pluginscope["decision_variables"]
            ]

        normalize_par1 = self.normalize(var_values[0], min_par1, max_par1)
        # normalize_par2 = self.normalize(var_values[1], min_par2, max_par2)

        diff_par1 = abs(gt_par1 - normalize_par1)
        # diff_par2 = abs(gt_par2 - normalize_par2)

        # Normalize together
        score = self.normalize(
            diff_par1, 0, len(self.pluginscope["decision_variables"])
        ) #(+diff_par_2)

        # The higher the score, the closer it is to the real parameter
        return 1 - score

    def normalize(self, x, min, max):
        return (x - min) / (max - min)

    def get_summary(self):
        df_solutions = pd.DataFrame.from_dict(
            self.pluginscope["solutions"], orient="index"
        ).drop_duplicates()
        dict_decision_variables = df_solutions.iloc[:, :len(self.pluginscope["decision_variables"])].to_dict("records")[0]
        #TODO check if this works with multiprocessing
        return dict_decision_variables, f"Summary CalibrationModel with solutions \n {df_solutions}" \
                                        f"\n and eprogress is {self.pluginscope['eprogress']}"

    def get_score(self):
        return self.pluginscope["score"]

    def get_report(self, **kwargs) -> CalibrationModelReport:
        report = GeneticAlgorithmModelReport(self, **kwargs)
        return report
