import pandas as pd
import numpy as np
import random
import json

import gurobipy as gp
from gurobipy import GRB

from celibration import (
    CalibrationModel,
    CalibrationModelFactory,
    CalibrationDifferenceFunction,
    CalibrationModelReport,
)
from typing import Any

from stylized_supply_chain_model.RunnerObject import (
    SupplyChainSimModel,
)


class PluginFactory(CalibrationModelFactory):
    def create_calibration_model(self, info=dict) -> CalibrationModel:
        return GurobiModel(info=info)


class GurobiReport(CalibrationModelReport):
    # self._model = model is set in the default constructor
    def get_string(self) -> str:
        result = (
            f"Report:\033[1m GurobiReport \033[0m:\n"
            # f"\tParameters:\n"
            # f"\t\t{self._kwargs}\n"
            f"\tModel:\n"
            f"\t\tResults: {self._model.get_summary()[1]}\n"
        ).expandtabs(2)
        return result

    def render(self, type: str) -> Any:
        return NotImplementedError

    def export_to_file(self, type: str, filename: str) -> Any:
        if type == "pdf":
            self.export_to_pdf(filename=filename)
        if type == "json":
            self.export_to_json(filename=filename)
        return NotImplementedError

    def export_to_json(self, filename: str) -> Any:
        cal_model_json = {
            "Model_info": self._model.get_info(),
            "Results": self._model.pluginscope["solutions"].to_dict(),
        }
        with open(filename, "w", encoding="utf-8") as f:
            f.write(json.dumps(cal_model_json))


class GurobiModel(CalibrationModel):
    def fit(
            self,
            df_in: pd.DataFrame,
            diff_func_object: CalibrationDifferenceFunction,
            diff_func_parameters: dict,
            debug: bool,
            **kwargs,
    ):
        """Performs the calibration via the Bayesian Optimization. This will be 
        further elaborated in the documentation per functions. 

        Args:
            df_in (Dataframe): dataframe to fit on
            diff_func_object (object): difference function to use as likelihood
            kwargs: arguments in key-value format
        """
        if "seed" in kwargs:
            random.seed(kwargs["seed"])
            np.random.seed(kwargs["seed"])

        self.pluginscope["df_ground_truth"] = df_in

        self.pluginscope["decision_variables"] = kwargs["decision_variables"]
        self.pluginscope["ranges_variables"] = kwargs["ranges_variables"]
        self.pluginscope["objectives"] = kwargs["objectives"]

        results = self.gurobi_model()

        df_solutions, objective = self.get_solutions(results)
        self.pluginscope["score"] = self.calculate_score(df_solutions, self.pluginscope["ranges_variables"])  # score
        self.pluginscope["solutions"] = df_solutions

        return

    def gurobi_model(self):

        model = gp.Model("opt_supply_chain")

        lower_bound = self.pluginscope["ranges_variables"][0][0]
        upper_bound = self.pluginscope["ranges_variables"][0][1]

        var1 = model.addVar(lb=lower_bound, ub=upper_bound, vtype=GRB.CONTINUOUS,
                            name=self.pluginscope["decision_variables"][0])
        #par1.Start = np.random.randint(lower_bound, upper_bound)
        model.update()

        obj = self.test_distance_function  # it goes wrong that we already "call" the function before optimization

        def test():
            return gp.LinExpr(obj(var1))

        print("set objective")

        model.setObjective(test(), GRB.MINIMIZE)

        print("optimization starts")
        model.optimize()

        return model

    def test_distance_function(self, par1):
        return 1*self.distance_function(par1)

    def distance_function(self, par1):
        """Calculates the distance for one or more objectives, based on the 
        results of the simulation model and difference function. The results of
        the simulation model are calculated with the given parameters.
        If there is more than one objective, the distances between the df_in and
        the results of the simulation model are summed.

        Args:
            par1 (float): Value of parameter 1

        Returns:
            float: Returns the value of the distance, which is the objective.
        """
        # Run simulation
        par1 = par1
        # par2 = round(par2)

        result_sim_model = SupplyChainSimModel.run(parameters=[par1])

        # Calculate distance metrics
        if len(self.pluginscope["objectives"]) == 1:
            print("One objective")
            dist = self.calculate_objectives(result_sim_model)[0]

        elif len(self.pluginscope["objectives"]) > 1:
            print("More objectives but mono-objective")
            all_obj_dist = self.calculate_objectives(result_sim_model)
            dist = sum(all_obj_dist)

        return dist

    def calculate_objectives(self, result_sim_model):
        """Calculates the distance per parameter between the ground truth and the results of 
        the simulation model, which is the objective. The distance is calculated by the given 
        difference function(s).

        Args:
            result_sim_model (Dataframe): Results of the simulation model

        Returns:
            list: Distance per parameter, which is objective
        """
        obj_dist = []
        for obj in self.pluginscope["objectives"]:
            dist = self.diff_func_object.calculate(
                self.pluginscope["df_ground_truth"][obj],
                result_sim_model[obj],
                debug=False,
            )
            # print(
            #     "Solution: {0} {1:.2f} and {2} {3} with {4} gives {5}".format(
            #         self.pluginscope["decision_variables"][0],
            #         par1,
            #         self.pluginscope["decision_variables"][1],
            #         par2,
            #         obj,
            #         -dist,
            #     )
            # )
            obj_dist.append(dist)
        return obj_dist

    def get_solutions(self, results):
        """The results are received by the bayesion optimisation.The score is the maximum result 
        of the BO model, where the negative is taken because the BO can only maximise. The values 
        of the parameters corresponding to the maximum results are converted into the solution 
        dataframe. Parameter 2 is converted to an integer. The bayesian optimisation gives the 
        solution that will approach the global maximum.

        Args:
            results (dict): Results of bayesian optimisation

        Returns:
            dataframe: Associated parameters with which the bayesian optimisation is optimised.
            float: The minimum distance, which indicates how well the optimisation model performs.

        """
        # df_all_results = pd.DataFrame.from_dict(results.res)
        results_parameters = []

        for key, value in results.max["params"].items():
            if key == "par2":
                results_parameters.append(round(value))
            else:
                results_parameters.append(value)

        # Reverse -1*dist
        score = -results.max["target"]

        df_results = pd.DataFrame(
            [results_parameters], columns=self.pluginscope["decision_variables"]
        )
        df_score = pd.DataFrame([score], columns=["Score"])
        df_solutions = (
            df_results.join(df_score).drop_duplicates().reset_index(drop=True)
        )

        return df_solutions, score

    def calculate_score(self, solution, ranges):
        # TODO generiek maken, ground truth parameters & aantal parameters
        """How close to the ground truth parameters"""
        min_par1 = ranges[0][0]
        max_par1 = ranges[0][1]
        # min_par2 = ranges[1][0]
        # max_par2 = ranges[1][1]

        gt_par1 = self.normalize(2.5, min_par1, max_par1)
        # gt_par2 = self.normalize(1, min_par2, max_par2)

        # Normalize the difference
        var_values = [solution[var][0] for var in self.pluginscope["decision_variables"]]
        normalize_par1 = self.normalize(var_values[0], min_par1, max_par1)
        # normalize_par2 = self.normalize(var_values[1], min_par2, max_par2)

        diff_par1 = abs(gt_par1 - normalize_par1)
        # diff_par2 = abs(gt_par2 - normalize_par2)

        # Normalize together
        score = self.normalize(diff_par1, 0, len(self.pluginscope["decision_variables"]))
        # (+diff_par_2)

        # The higher the score, the closer it is to the real parameter
        return 1 - score

    def normalize(self, x, min, max):
        return (x - min) / (max - min)

    def get_summary(self):
        dict_decision_variables = \
        self.pluginscope["solutions"].iloc[:, :len(self.pluginscope["decision_variables"])].to_dict("records")[0]
        df_results = self.pluginscope["solutions"]
        return dict_decision_variables, f"Summary CalibrationModel with most optimal solution:  \n {df_results}"

    def get_score(self):
        return self.pluginscope["score"]

    def get_report(self, **kwargs) -> CalibrationModelReport:
        report = BayesianOptimizationReport(self, **kwargs)
        return report
