"""
Created on: 16-7-2021 10:32

@author: IvS
"""
import itertools

from pydsol.basic_logger import get_module_logger
import logging
from pydsol.pydsol_core.ExperimentReplication import ExperimentReplication

logger = get_module_logger(__name__, level=logging.INFO)


class Experiment(object):
    """ This class defines an experiment with multiple replications. Pydsol has one simulator per replication to
    guarantee a total reset of statistics. DSOL in Java has one simulator per experiment, and clears the
    eventlist per replication."""
    id_iter = itertools.count(1)

    def __init__(self, name_simulator, model, start_time, warmup_time, run_length, num_replications):
        """

        Parameters
        ----------
        name_simulator: class
            class of simulator, for example Simulator.
        model: DSOLModel
            DSOLModel object to experiment with.
        start_time: int
            the start time as a time object.
        warmup_time: int
            the warmup period, included in the runlength.
        run_length: int
            the total length of the run, including the warm-up period.
        num_replications: int
            number of replications to execute for one experiment.
        """
        self.id = next(self.id_iter)

        self.name_simulator = name_simulator
        if not hasattr(self.name_simulator, "__name__"):
            raise AttributeError("{0} is an instance, should be a Class".format(self.name_simulator))
        self.model = model

        # Run controls
        self.start_time = start_time
        self.warmup_time = warmup_time
        self.run_length = run_length
        self.num_replications = num_replications

        self.current_num_of_replications = 0

        self.summary_statistics = {}

    def start(self):
        """Start the entire experiment on the simulator, and execute the replications one by one. """
        while self.current_num_of_replications < self.num_replications:
            # try:
            self.start_next_replication()
            # except ValueError:
            #     raise ValueError("There is no replication.")
        self.end_experiment()

    def start_next_replication(self):
        """Start the next replication from the list of replications."""
        self.current_num_of_replications += 1

        replication = ExperimentReplication(self.current_num_of_replications, self.start_time,
                                            self.warmup_time, self.run_length, self)
        self.model.seed = self.update_seed(self.current_num_of_replications)

        simulator = self.name_simulator()
        self.model.simulator = simulator
        logger.debug("Model has Simulator {0}".format(simulator.id))
        simulator.initialize(self.model, replication)
        simulator.start()
        self.end_replication()

    # TODO this must be modified based on the user (?)
    def end_replication(self):
        """Create or update summary statistics for the experiment based on the statistics of the just completed
        replication. The summary statistics of the experiment are saved in a dictionary with as key the replication."""
        logger.debug(
            "Replications {0} of Experiment {1} has finished".format(self.current_num_of_replications, self.id))
        self.summary_statistics[self.current_num_of_replications] = self.model.get_output_statistics()

    def end_experiment(self):
        """Returns the output statistics of the experiment, so for each replication or combined. This must be
        defined by the user."""
        # Return output statistics for each replication or combined - depending on what you want
        logger.info("Experiment {0} with {1} replications is finished ".format(self.id, self.num_replications))
        logger.debug("Output statistics are {0}".format(self.summary_statistics))
        return self.summary_statistics

    def update_seed(self, current_rep_nr):
        """Updates the seed for each replication by adding the current model seed with the current replication number.

        Parameters
        ----------
        current_rep_nr: int
            current replication number.
        """
        seed = self.model.seed + current_rep_nr
        logger.debug("Replication nr {0} has seed {1}".format(current_rep_nr, seed))
        return seed
