"""
Created on: 25-6-2021 13:51

@author: IvS
"""
import itertools

from pydsol.basic_logger import get_module_logger
import logging
from pydsol.pydsol_core.EventList import EventList
from pydsol.pydsol_core.SimEvent import SimEvent


logger = get_module_logger(__name__, level=logging.INFO)


class Simulator(object):
    """This simulator defines a discrete event simulator. Each simulator has a new eventlist. """
    id_iter = itertools.count(1)

    def __init__(self):
        self.id = next(self.id_iter)

        self.run_time = 0
        self.warmup_time = 0
        self.simulation_time = 0  # GetStartTime of simulation (when there is warm-up time)

        self.eventlist = EventList()

        self.is_running = False
        is_stopping = not self.is_running  # heeft met multi-threading te maken

        self.model = object()
        self.replication = object()

    def initialize(self, model, replication):
        """Initialize the model and the replication by calling construct_model()

        Parameters
        ----------
        model: DSOLModel
            DSOLModel object
        replication: Replication
            Replication object
        """
        self.replication = replication
        self.run_time = self.replication.run_length
        self.warmup_time = self.replication.warmup_time
        self.simulation_time = self.replication.start_time

        self.eventlist.clear()
        self.model = model
        self.model.construct_model()

    def warm_up(self, source, target, method, priority=None, **kwargs):
        """Schedule warm up event to clear the statistics after warm up period. Warm up period
        is included in the run time by the user.

        Parameters
        ----------
        source: object
            source that creates the method.
        target:  object, or class for __init__
            target the target on which a state change is scheduled.
        method: str
            method is the method which embodies the state changes.
        priority: int
            priority of the event, default is normal priority.
        kwargs:
            kwargs are the keyword arguments that are used to invoke the method with."""

        warm_up_event = self.create_simevent(self.warmup_time, source, target, method, priority, **kwargs)
        self.schedule_event(warm_up_event)

    def schedule_event(self, event):
        """ Schedules a SimEvent by adding it to the eventlist.

        Parameters
        ----------
        event: SimEvent object

        """
        if event.absolute_execution_time < self.simulation_time:
            raise ValueError("{0} cannot be scheduled in the past".format(event))
        self.eventlist.add(event)
        logger.debug("Event {0} is scheduled at Time {1}".format(event.id, event.absolute_execution_time))

    def create_simevent(self, relative_delay, source, target, method, priority=None, **kwargs):
        """Creates a SimEvent with delayed method invocation.

        Parameters
        ----------
        relative_delay: int
            relative delay of the event (compared to the current simulation time).
        source: object
            source that creates the method.
        target:  object, or class for __init__
            target the target on which a state change is scheduled.
        method: str
            method is the method which embodies the state changes.
        priority: int
            priority of the event, default is normal priority.
        kwargs:
            kwargs are the keyword arguments that are used to invoke the method with.

        Returns
        -------
        object
            SimEvent
        """
        abs_eventtime = self.simulation_time
        abs_eventtime += relative_delay
        logger.debug("New SimEvent with Time {0:.2f} is created".format(abs_eventtime))
        return SimEvent(abs_eventtime, source, target, method, priority, **kwargs)

    def create_simevent_now(self, source, target, method, priority=None, **kwargs):
        """Creates SimEvent for now, meaning the relative delay is 0.

        Parameters
        ----------
        source: object
            source that creates the method.
        target:  object, or class for __init__
            target the target on which a state change is scheduled.
        method: str
            method is the method which embodies the state changes.
        priority: int
            priority of the event, default is normal priority.
        kwargs:
            kwargs are the keyword arguments that are used to invoke the method with.
        Returns
        -------
        object
            SimEvent
        """
        logger.debug("New SimEvent with Time {0:.2f} is created for now".format(self.simulation_time))
        return SimEvent(self.simulation_time, source, target, method, priority, **kwargs)

    def start(self):
        """
        Starts the simulation by repeatedly removing the first event from the eventlist and executing this event,
        until the eventlist is empty or the run time is completed.
        """
        self.is_running = True
        while self.is_running:
            if self.eventlist.is_empty():
                self.is_running = False
                logger.debug("Event list is empty")
                break

            event = self.eventlist.remove_first()

            if self.run_time < event.absolute_execution_time:
                self.is_running = False
                logger.debug("First event is not within run time")
                break

            self.simulation_time = event.absolute_execution_time
            logger.debug("Time {0:.2f}: Simulation Time".format(self.simulation_time))

            # try:
                # Execute events
            event.execute()

            # except Exception:
            #     # Stopped the simulation
            #     self.is_running = False
            #     raise Exception
