"""
Created on: 12-8-2021 10:59

@author: IvS
"""
import itertools
import numpy as np

from pydsol.pydsol_model.queue import Queue
from pydsol.pydsol_model.node import Node

import logging
from pydsol.basic_logger import get_module_logger

logger = get_module_logger(__name__, level=logging.DEBUG)


class ServerWithCapacity(object):
    id_iter = itertools.count(1)

    def __init__(self, simulator, capacity=1, processing_time="default", transfer_in_time=0, **kwargs):
        self.simulator = simulator

        self.id = next(self.id_iter)
        self.name = "{0} {1}".format(self.__class__.__name__, str(self.id))
        if "name" in kwargs:
            self.name = kwargs["name"]

        self.processing_time = processing_time

        self.capacity = capacity
        self.transfer_in_time = transfer_in_time
        self.kwargs = kwargs

        # Transport
        self.input_links = []
        self.output_links = []

        self.input_node = Node(self.simulator)
        self.output_node = Node(self.simulator)

        # Queue
        self.input_queue = Queue(self)
        self.resources = {}
        for cap in range(self.capacity):
            self.resources[cap] = False

    def enter_input_node(self, entity):
        event = self.simulator.create_simevent_now(self, entity, "seize_resource")
        self.simulator.schedule_event(event)

    def seize_resource(self, entity, **kwargs):
        """Process to seize a resource by the entity. If the resource is free and there is no queue, the entity
        can directly seize the resource. If the resource is free and there is a queue, the first entity of the queue
        seizes the resource and this entity is added to the queue. If the resource is occupied, the entity is directly
        added to the queue.

        Parameters
        ----------
        entity: Entity
            the target on which a state change is scheduled.
        kwargs:
            kwargs are the keyword arguments that are used to expand the function.

        """
        not_seized = [key for key, value in self.resources.items() if not value]

        if len(not_seized) == 0:
            self.input_queue.contents.append(entity)
            logger.info(
                "Time {0:.2f}: {1} added to the queue of {2} with length {3}".format(
                    self.simulator.simulation_time, entity.name,
                    self.name, len(self.input_queue.contents)))
        else:
            resource_to_seize = not_seized[0]
            if len(self.input_queue.contents) == 0:
                self.resources[resource_to_seize] = True
                event = self.simulator.create_simevent_now(self, entity, "exit_input_node", resource=resource_to_seize)
                self.simulator.schedule_event(event)
                logger.debug(
                    "Time {0:.2f}: {1} seized Resource {2} at {3}".format(self.simulator.simulation_time,
                                                                          entity.name, resource_to_seize, self.name))
            else:
                self.resources[resource_to_seize] = True
                event = self.simulator.create_simevent_now(self, self.input_queue.contents[0], "exit_input_node",
                                                           resource=resource_to_seize)
                self.simulator.schedule_event(event)
                logger.debug(
                    "Time {0:.2f}: {1} seized Resource {2} at {3}".format(
                        self.simulator.simulation_time, self.input_queue.contents[0].name,
                        resource_to_seize, self.name))
                self.input_queue.contents.pop(0)
                self.input_queue.contents.append(entity)

    def enter_resource(self, entity, **kwargs):
        event = self.simulator.create_simevent(self.transfer_in_time, self, entity, "processing",
                                               resource=kwargs["resource"])
        self.simulator.schedule_event(event)

    def processing(self, entity, **kwargs):
        if self.processing_time == "default":
            self.processing_time = np.random.triangular(0.1, 0.2, 0.3)  # minutes

        event = self.simulator.create_simevent(self.processing_time, self, entity, "exit_resource",
                                               resource=kwargs["resource"])
        self.simulator.schedule_event(event)

    def exit_resource(self, entity, **kwargs):
        # Enter link for entity
        try:
            logger.info(
                "Time {0:.2f}: {1} is processed by {2}".format(self.simulator.simulation_time, entity.name, self.name))
            weights = [link.selection_weight for link in self.output_links]
            link_by_weight = np.random.choice(np.array(self.output_links), p=weights / np.sum(weights))
            link_by_weight.enter_link(entity)

        except AttributeError:
            raise AttributeError("{0} has no output link".format(self.name))

        # Release resource and check queue
        self.resources[kwargs["resource"]] = False
        event = self.simulator.create_simevent_now(self, self, "check_queue", resource=kwargs["resource"])
        self.simulator.schedule_event(event)

    def check_queue(self, **kwargs):
        logger.info(
            "Time {0:.2f}: Length of queue attached to {1} is {2}".format(self.simulator.simulation_time, self.name,
                                                                          len(self.input_queue.contents)))
        if len(self.input_queue.contents) == 0:
            pass
        else:
            self.resources[kwargs["resource"]] = True
            event = self.simulator.create_simevent_now(self, self.input_queue.contents[0], "enter_resource",
                                                       resource=kwargs["resource"])
            self.simulator.schedule_event(event)
            logger.debug(
                "Time {0:.2f}: {1} seized Resource {2} at {3}".format(self.simulator.simulation_time,
                                                                      self.input_queue.contents[0].name,
                                                                      kwargs["resource"],
                                                                      self.name))
            self.input_queue.contents.pop(0)
