"""
Created on: 22-7-2021 10:19

@author: IvS
"""
import logging

import numpy as np
import itertools

from pydsol.pydsol_model.node import Node

from pydsol.basic_logger import get_module_logger

logger = get_module_logger(__name__)


class Source(object):
    """This class defines a basic source for a discrete event simulation model. A Source is the start-station
    of each entity, meaning the entities enter the system via the source. Entities are created at a source,
    given a specific interarrival time."""
    id_iter = itertools.count(1)

    def __init__(self, simulator, interarrival_time="default", num_entities=1, **kwargs):
        """

        Parameters
        ----------
        simulator: Simulator object
            simulator of the model.
        interarrival_time:
            time between the creation of entities. Default is np.random.exponential(0.25).
        num_entities: int
            number of entities to create at once. Default is 1.
        kwargs:
            kwargs are the keyword arguments that are used to expand the source class.
            *name: str
                user-specified source for the link
        """
        self.simulator = simulator
        self.num_entities = num_entities

        self.id = next(self.id_iter)
        self.name = "{0} {1}".format(self.__class__.__name__, str(self.id))
        if "name" in kwargs:
            self.name = kwargs["name"]

        self.output_links = []

        self.interarrival_time = interarrival_time
        if "distribution" in kwargs:
            self.distribution = kwargs["distribution"]

        self.output_node = Node(self.simulator)

        self.kwargs = kwargs

    def create_entities(self, entity_type, **kwargs):
        """
        Create entities via SimEvent, given the interarrival time and the number of entities.

        Parameters
        ----------
        entity_type: class
            class where to make instances of, for example class Entity
        kwargs:
            kwargs are the keyword arguments that are used to invoke the method or expand the function.

        """

        # To make it work with the set seed of the simulator
        if self.interarrival_time == "default":
            self.interarrival_time = np.random.exponential(0.25)

        # Create new entity
        for _ in range(self.num_entities):
            event = self.simulator.create_simevent(self.simulator.simulation_time, self, entity_type,
                                                   "__init__", **kwargs, simulator=self.simulator,
                                                   source_entity=self)
            self.simulator.schedule_event(event)

        interarrival_time = self.distribution(
            self.interarrival_time) if "distribution" in self.__dict__ else self.interarrival_time
        relative_delay = interarrival_time

        # Schedule event to create next entity according to the interarrival time
        new_event = self.simulator.create_simevent(relative_delay, self, entity_type, "create_entities", **kwargs)
        self.simulator.schedule_event(new_event)

    def exit_source(self, entity, **kwargs):
        """Schedules the event to exit the source and enter the output node.

        Parameters
        ----------
        entity: object
            the target on which a state change is scheduled.
        kwargs:
            kwargs are the keyword arguments that are used to expand the function.

        """
        logger.debug("Time {0:.2f}: {1} exists {2}".format(self.simulator.simulation_time, entity.name, self.name))
        event = self.simulator.create_simevent_now(self, entity, "enter_output_node")
        self.simulator.schedule_event(event)

    def enter_output_node(self, entity, **kwargs):
        """Combine the entity with a Vehicle if an vehicle type is given. Combined or not, it schedules an event
        for exiting the output node.

        Parameters
        ----------
        entity: object
            the target on which a state change is scheduled.
        kwargs:
            kwargs are the keyword arguments that are used to expand the function.
            *vehicle_type: Vehicle class
                subclass of Vehicle on which the entity should travel to the next destination.
            *vehicle_speed: int, optional
                speed of the vehicle.

        """
        if "vehicle_type" in self.kwargs:
            if "vehicle_speed" in self.kwargs:
                vehicle = self.kwargs["vehicle_type"](self.simulator, self.kwargs["vehicle_speed"])
            else:
                vehicle = self.kwargs["vehicle_type"](self.simulator)
            vehicle.entities_on_vehicle.append(entity)
            logger.info("Time {0:.2f}: {1} loaded on {2}".format(self.simulator.simulation_time, entity.name,
                                                                 vehicle.name))
            entity = vehicle

        logger.debug("Time {0:.2f}: {1} enters {2}".format(self.simulator.simulation_time, entity.name, self.output_node.name))
        event = self.simulator.create_simevent_now(self, entity, "exit_output_node")
        self.simulator.schedule_event(event)

    def exit_output_node(self, entity, **kwargs):
        """Exit the resource by selecting a link on which the entity should travel to the next destination by
        weighted choice.

        Parameters
        ----------
        entity: object
            the target on which a state change is scheduled.
        kwargs:
            kwargs are the keyword arguments that are used to expand the function.

        """
        logger.debug("Time {0:.2f}: Exit Source of {1}".format(self.simulator.simulation_time, entity.name))

        try:
            weights = [link.selection_weight for link in self.output_links]
            link_by_weight = np.random.choice(np.array(self.output_links), p=weights / np.sum(weights))
            link_by_weight.enter_link(entity)
        except AttributeError:
            raise AttributeError("{0} has no output link".format(self.name))





