"""
Created on: 6-8-2021 10:09

@author: IvS
"""
import itertools
import logging
import statistics
import numpy as np
import pandas as pd

from pydsol.pydsol_core.DSOLModel import DSOLModel

from stylized_supply_chain_model.model_elements.product import Product
from stylized_supply_chain_model.model_elements.supplier import Supplier
from stylized_supply_chain_model.model_elements.manufacturer import Manufacturer
from stylized_supply_chain_model.model_elements.port import Port
from stylized_supply_chain_model.model_elements.wholesales_distributor import WholesalesDistributor
from stylized_supply_chain_model.model_elements.retailer import Retailer
from pydsol.pydsol_model.link import Link
from stylized_supply_chain_model.model_elements.link_sea import SeaLink

from stylized_supply_chain_model.model_elements.vehicles.large_truck import LargeTruck
from stylized_supply_chain_model.model_elements.vehicles.small_truck import SmallTruck
from stylized_supply_chain_model.model_elements.vehicles.feeder import Feeder
from stylized_supply_chain_model.model_elements.vehicles.vessel import Vessel

from stylized_supply_chain_model.model_elements.timer import Timer

from pydsol.basic_logger import get_module_logger

logger = get_module_logger(__name__)


class SimModel(DSOLModel):
    def __init__(self, input_parameters, **kwargs):
        super().__init__(input_parameters, **kwargs)
        self.input_parameters = input_parameters

        self.supplier = None
        self.manufacturer = None
        self.export_port = None
        self.transit_port = None
        self.import_port = None
        self.wholesales_distributor = None
        self.retailer_amsterdam = None
        self.retailer_utrecht = None
        self.retailer_venlo = None

    def construct_model(self):
        """Model has the time unit km/h"""
        # Reset model
        self.reset_model()

        np.random.seed(self.seed)

        # Set up model
        supplier = Supplier(self.simulator, interarrival_time=1/15,
                            location="Vietnam",
                            distribution=np.random.exponential,
                            vehicle_type=SmallTruck)
        manufacturer = Manufacturer(self.simulator, location='Vietnam',
                                    processing_time=self.input_parameters["manufacturing_time"],
                                    vehicle_type=LargeTruck)

        inland_carrier_1 = Link(self.simulator, supplier, manufacturer, 50)

        export_port = Port(self.simulator, location='Hai Phong', transfer_in_time=(0, 0.2, 0.5),
                           processing_time=(2, 2, 1), vehicle_type=Feeder)

        inland_carrier_2 = Link(self.simulator, manufacturer, export_port, 125)

        transit_port = Port(self.simulator, location='Tanjung Pelepas', transfer_in_time=(0, 0.2, 0.5),
                            processing_time=(2, 2, 1), vehicle_type=Vessel)

        sea_carrier_feeder = SeaLink(self.simulator, export_port, transit_port, 1656)  # these are nautical miles

        import_port = Port(self.simulator, location='Rotterdam', transfer_in_time=(0, 0.2, 0.5),
                           processing_time=(2, 2, 1), vehicle_type=LargeTruck)

        sea_carrier_vessel = SeaLink(self.simulator, transit_port, import_port, 9286)  # these are nautical miles

        wholesales_distributor = WholesalesDistributor(self.simulator,
                                                       minimum_processing_time=0.5,
                                                       mode_processing_time=1,
                                                       maximum_processing_time=2,
                                                       location='Eindhoven', vehicle_type=SmallTruck,
                                                       divide_quantity=3)

        inland_carrier_3 = Link(self.simulator, import_port, wholesales_distributor, 135)

        retailer_amsterdam = Retailer(self.simulator, transfer_in_time=0.2, distribution=np.random.exponential,
                                      location='Amsterdam')
        retailer_utrecht = Retailer(self.simulator, transfer_in_time=0.2, location='Utrecht')
        retailer_venlo = Retailer(self.simulator, transfer_in_time=0.2, location='Venlo')

        inland_carrier_5 = Link(self.simulator, wholesales_distributor, retailer_amsterdam, 125)
        inland_carrier_6 = Link(self.simulator, wholesales_distributor, retailer_utrecht, 92)
        inland_carrier_7 = Link(self.simulator, wholesales_distributor, retailer_venlo, 60)

        # Set up stats
        self.supplier = supplier
        self.manufacturer = manufacturer
        self.export_port = export_port
        self.transit_port = transit_port
        self.import_port = import_port
        self.wholesales_distributor = wholesales_distributor
        self.retailer_amsterdam = retailer_amsterdam
        self.retailer_utrecht = retailer_utrecht
        self.retailer_venlo = retailer_venlo

        self.set_output_stats()

        supplier.create_entities(Product, interarrival_time=self.input_parameters["manufacturing_time"])

    @staticmethod
    def reset_model():
        Product.id_iter = itertools.count(1)
        Supplier.id_iter = itertools.count(1)
        Manufacturer.id_iter = itertools.count(1)
        Port.id_iter = itertools.count(1)
        WholesalesDistributor.id_iter = itertools.count(1)
        Retailer.id_iter = itertools.count(1)

        Link.id_iter = itertools.count(1)
        Feeder.id_iter = itertools.count(1)
        Vessel.id_iter = itertools.count(1)
        SmallTruck.id_iter = itertools.count(1)
        LargeTruck.id_iter = itertools.count(1)

    def set_output_stats(self):
        timer = Timer(self.simulator, 1)
        timer.set_event(self.supplier, self.supplier, "get_hourly_stats")
        timer.set_event(self.manufacturer, self.manufacturer, "get_hourly_stats")
        timer.set_event(self.export_port, self.export_port, "get_hourly_stats")
        timer.set_event(self.transit_port, self.transit_port, "get_hourly_stats")
        timer.set_event(self.import_port, self.import_port, "get_hourly_stats")
        timer.set_event(self.wholesales_distributor, self.wholesales_distributor,
                        "get_hourly_stats")
        timer.set_event(self.retailer_amsterdam, self.retailer_amsterdam, "get_hourly_stats")
        timer.set_event(self.retailer_utrecht, self.retailer_utrecht, "get_hourly_stats")
        timer.set_event(self.retailer_venlo, self.retailer_venlo, "get_hourly_stats")

    def get_output_statistics(self):
        # Average product time in system
        average_product_time_in_system = statistics.mean(
            [product.time_in_system for product in self.retailer_amsterdam.entities_of_system] +
            [product.time_in_system for product in self.retailer_utrecht.entities_of_system] +
            [product.time_in_system for product in self.retailer_venlo.entities_of_system])
        # Average international transport time
        average_international_transport_time = statistics.mean(
            [product.international_transport_time for product in self.retailer_amsterdam.entities_of_system] +
            [product.international_transport_time for product in self.retailer_utrecht.entities_of_system] +
            [product.international_transport_time for product in self.retailer_venlo.entities_of_system])
        # Average wholesales distributor time
        average_wholesales_distributor_time = statistics.mean(
            [product.wholesales_distributor_time for product in self.retailer_amsterdam.entities_of_system] +
            [product.wholesales_distributor_time for product in self.retailer_utrecht.entities_of_system] +
            [product.wholesales_distributor_time for product in self.retailer_venlo.entities_of_system])
        # Average quantity
        average_quantity = statistics.mean(self.supplier.list_quantity)

        outcomes = {"Time_In_System": average_product_time_in_system,
                    "International_Transport_Time": average_international_transport_time,
                    "Wholesales_Time": average_wholesales_distributor_time,
                    "Quantity": average_quantity}

        ## Time series as dictionary
        # time_series = {"Supplier": self.supplier.hour_stats,
        #                "Manufacturer": self.manufacturer.hour_stats,
        #                "Export_Port": self.export_port.hour_stats,
        #                "Transit_Port": self.transit_port.hour_stats,
        #                "Import_port": self.import_port.hour_stats,
        #                "Wholesales_Distributor": self.wholesales_distributor.hour_stats,
        #                "Retailer_Amsterdam": self.retailer_amsterdam.hour_stats,
        #                "Retailer_Utrecht": self.retailer_utrecht.hour_stats,
        #                "Retailer_Venlo": self.retailer_venlo.hour_stats}

        time_series = pd.DataFrame({"Supplier": pd.Series(self.supplier.hour_stats),
                                    "Manufacturer": pd.Series(self.manufacturer.hour_stats),
                                    "Export_Port": pd.Series(self.export_port.hour_stats),
                                    "Transit_Port": pd.Series(self.transit_port.hour_stats),
                                    "Import_Port": pd.Series(self.import_port.hour_stats),
                                    "Wholesales_Distributor": pd.Series(
                                        self.wholesales_distributor.hour_stats),
                                    "Retailer_Amsterdam": pd.Series(self.retailer_amsterdam.hour_stats),
                                    "Retailer_Utrecht": pd.Series(self.retailer_utrecht.hour_stats),
                                    "Retailer_Venlo": pd.Series(self.retailer_venlo.hour_stats)})
        time_series.reset_index(inplace=True)
        time_series = time_series.rename(columns={"index": "Time"})

        result = {"outcomes": outcomes, "time_series": time_series}

        return result
