"""
Created on: 6-8-2021 10:09

@author: IvS
"""
import time
from datetime import date
import pandas as pd

from stylized_supply_chain_model.model import SimModel
from pydsol.pydsol_core.Experiment import Experiment
from pydsol.pydsol_core.Simulator import Simulator

from ema_workbench import Model, IntegerParameter, RealParameter, TimeSeriesOutcome, SequentialEvaluator

from pydsol.basic_logger import get_module_logger
import logging

logger = get_module_logger(__name__, level=logging.DEBUG)


def run_simulation(manufacturing_time=2.5, replications=5):
    """
    This function runs the simulation model with user-defined interarrival time and route via rotterdam,
    with user-defined number of replications. It saves the timeseries to a CSV file.
vo
    Parameters
    ----------
    manufacturing_time: Union[int, float]
        inputparameter for the manufacturing time of the product. Default is 2.5 days.
    replications: int
        number of replications. Default is 5.

    Returns
    -------
    dict
        dictionary with the timeseries in a pandas DataFrame


    """
    sim_model = SimModel(input_parameters={"manufacturing_time": manufacturing_time})

    experiment = Experiment(Simulator, sim_model, 0, 0, 7 * 10, replications) #10 weeks
    experiment.start()
    all_results = experiment.summary_statistics

    all_timeseries_df = []
    # Add replication number
    for key in experiment.summary_statistics.keys():
        time_series = experiment.summary_statistics[key]["time_series"]
        time_series["Replications"] = key

        all_timeseries_df.append(time_series)

    # Combine all time series
    all_timeseries = pd.concat(all_timeseries_df).reset_index(drop=True)
    # # Save as CSV
    # date_today = date.today().strftime("%Y%m%d")
    # all_timeseries.to_csv(
    #     "./data/" + date_today + "_GT_TimeSeries_ManufacturingTime_" + str(round(manufacturing_time, 2)) + ".csv")

    # TODO Split out to outcomes and time series
    return {"time_series": all_timeseries}


if __name__ == "__main__":
    start_time = time.time()

    run_once = True

    if run_once:
        # Run without uncertainty - only once
        results = run_simulation(replications=10)
        logger.info("Running takes {0:.1f} seconds".format((time.time() - start_time)))

    else:
        # Run with uncertainties - multiple scenarios
        model = Model("PhDSimModel", function=run_simulation)
        model.uncertainties = [RealParameter("manufacturing_time", 0.1, 10)]

        model.outcomes = [TimeSeriesOutcome("time_series")]

        nr_scenarios = 1
        with SequentialEvaluator(model) as evaluator:
            results = evaluator.perform_experiments(nr_scenarios)

        exp, outcome = results
        logger.info("Running {0} scenarios takes {1:.1f} seconds".format(nr_scenarios, (time.time() - start_time)))
